/**
 ** Original Code
 ** Copyright (c) 1998 Reid Barton, Dan Adkins, Harald Prokop,
 ** Copyright (c) 1998 Matteo Frigo, Chris Joerg, Martin Rinard,
 ** Copyright (c) 1998 Don Dailey, Charles E. Leiserson

 ** Permission is hereby granted, free of charge, to any person
 ** obtaining a copy of this software and associated documentation
 ** files (the "Software"), to deal in the Software without
 ** restriction, including without limitation the rights to use,
 ** copy, modify, merge, publish, distribute, sublicense, and/or sell
 ** copies of the Software, and to permit persons to whom the
 ** Software is furnished to do so, subject to the following
 ** conditions:

 ** The above copyright notice and this permission notice shall be
 ** included in all copies or substantial portions of the Software.

 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 ** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 ** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 ** NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 ** HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 ** WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 ** FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 ** OTHER DEALINGS IN THE SOFTWARE.

 ** 

 ** Port to Cilk++
 ** Copyright (c) 2010 Ruben Perez, Gregory Malecha

 ** Permission is hereby granted, free of charge, to any person
 ** obtaining a copy of this software and associated documentation
 ** files (the "Software"), to deal in the Software without
 ** restriction, including without limitation the rights to use,
 ** copy, modify, merge, publish, distribute, sublicense, and/or sell
 ** copies of the Software, and to permit persons to whom the
 ** Software is furnished to do so, subject to the following
 ** conditions:

 ** The above copyright notice and this permission notice shall be
 ** included in all copies or substantial portions of the Software.

 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 ** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 ** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 ** NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 ** HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 ** WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 ** FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 ** OTHER DEALINGS IN THE SOFTWARE.
 **/
/* Abort downward */
#include <strings.h>
class Abort {
  static const int GRAIN_SIZE = 8;

  struct Node {
    Node* next;
    int count;
    Abort* data[GRAIN_SIZE];
  };
  Node *children;
  bool aborted;
  Abort** parent;

  cilk::mutex m;

 private:
  void registerListener(Abort* res) {
    m.lock();

    res->aborted = this->aborted;
    if (this->aborted) {
      res->parent  = NULL;
    } else {
      if (this->children->count == GRAIN_SIZE) {
	Node* n = new Node();
	assert(n);

	n->count = 1;
	n->next = this->children;
	n->data[0] = res;
	bzero(n->data+1, (GRAIN_SIZE-1) * sizeof(Abort*));

	res->parent = &(n->data[0]);
      } else {
	this->children->data[this->children->count] = res;
	res->parent = &children->data[this->children->count++];
      }
    }
 
    m.unlock();
  }

  void propagate_abort() {
    if (this->aborted) return;

    m.lock();

    if (this->aborted) {
      m.unlock();
      return;
    }
    
    this->aborted = true;
    
    Node *children = this->children;
    while (children) {
      for (int i = 0; i < GRAIN_SIZE; i++) {
	Abort* r = children->data[i];
	if (!r) continue;

	if(__sync_bool_compare_and_swap(&(children->data[i]), r, NULL)) {
	  r->propagate_abort();
	} else {
	  assert(children->data[i] == NULL);
	}
      }	
      children = children->next;
    }

    m.unlock();
  }


 public:
  Abort() : aborted(false), parent(NULL) {
    //fprintf(stderr, "Abort() @ %p\n", this);
    this->children = new Node();
    assert(this->children);
    bzero(this->children, sizeof(Node));
  }

  Abort(Abort* parent) : children(NULL) {
    //fprintf(stderr, "Abort(%p) @ %p\n", parent, this);
    parent->registerListener(this);

    m.lock();
    if (this->aborted) {
      this->children = NULL;
    } else {
      this->children = new Node();
      assert(this->children);
      bzero(this->children, sizeof(Node));
    }
    m.unlock();
  }

  ~Abort() {

    if (this->parent) {
      if (__sync_bool_compare_and_swap(this->parent, this, NULL)) {
	// We haven't been aborted & we won't be called from above
      } else {
	while (!this->aborted);
	m.lock(); // Waits until the abort has finished
	m.unlock();
      }
    } else {
      // Won't be called from above
    }

    Node* t = this->children;
    while (t) {
      Node* r = t;
      t = r->next;
      delete r;
    }
  }

  inline int isAborted() const {
    return this->aborted;
  }

  void abort() {
    Abort** tmp = this->parent;
    if (tmp) {
      if (__sync_bool_compare_and_swap(tmp, this, NULL)) {
	// won't get any more calls from above
      } else {
	// an abort is going to get propagated from above...
	return;
      }
    }
    
    this->propagate_abort();
  }
};















#if 0
  Abort* serial_speculate() {
    Abort* r;
    if (this->aborted) {
      r = new Abort();
      r->aborted = true;
      return r;
    }

    r = new Abort(this);

    if (this->last == GRAIN_SIZE) {
      Node* n = new Node();
      n->data[0] = r;
      bzero(n->data+1, (GRAIN_SIZE-1) * sizeof(Abort*));
      n->next = this->children;
      this->last = 0;
      this->children = n;

      r->parent = &(n->data[0]);
    } else {
      this->children->data[this->last] = r;
      r->parent = &(this->children->data[this->last++]);
    }
     
    assert(*(r->parent) == r);
    r->aborted = this->aborted;
    if (r->aborted)
      *(r->parent) = NULL;
    return r;
  }

  Abort* speculate() {
    m.lock();
    Abort* r = serial_speculate();
    m.unlock();
    return r;
  }

  Abort* speculate() {
    /** Potential Race : under weak memory models **/
    Abort *r = new Abort();

    while (true) {
      int idx = ATOMIC_INCR(this->last);
      Abort* at = &this->children;
      
      if (idx < GRAIN_SIZE) {
	at->data[idx] = r;   // ATOMIC COPY
	r->parent = &(at->data[idx]);
	break;
      } else if (idx == GRAIN_SIZE) {
	Node* n = new Node();
	n->data[0] = r;
	bzero(n->data+1, (GRAIN_SIZE-1) * sizeof(Abort*));
	n->next = this->children;

	// BEGIN: SEQUENTIAL
	this->children = n; // ATOMIC COPY
	this->last = 1;
	// END: SEQUENTIAL
	break;
      } else {
	continue;
      }
    }

    r->aborted = this->aborted;
    return r;
  }
#endif
